/* SofChat (Web Chat program SHM version)      */
/* Copyright(c) 1996-1997 Y.YONEYAMA           */
/*                                             */
/* html.c                                      */

#include <stdio.h>
#include "html.h"
#include "messages.h"

int tagstack[256];
int tags;

char *validtags1[] = {
  "br",
  "dd",
  "dt",
  "hr",
/*  "img",*/
  "input",
  "li",
  "option",
  "tr",
  NULL
};
char *validtags2[] = {
  "a",
  "address",
  "b",
  "big",
  "blink",
  "blockquote",
  "caption",
  "center",
  "cite",
  "code",
  "dd",
  "dfn",
  "dir",
  "div",
  "dl",
  "em",
  "font",
  "form",
  "h1",
  "h2",
  "h3",
  "h4",
  "h5",
  "h6",
  "i",
  "kbd",
  "menu",
/*  "marquee",*/
  "nobr",
  "ol",
  "p",
  "pre",
  "s",
  "samp",
  "select",
  "small",
  "strike",
  "strong",
  "sub",
  "sup",
  "table",
  "textarea",
  "td",
  "th",
  "tr",
  "tt",
  "u",
  "ul",
  "var",
  NULL
};

int isvalidtag(tag)
char *tag;
{
  int i;
  for(i=0;validtags1[i] != NULL;i++)
  {
    if(strcmp(tag, validtags1[i]) == 0) return(0);
  }
  for(i=0;validtags2[i] != NULL;i++)
  {
    if(strcmp(tag, validtags2[i]) == 0) return(i+1);
  }
  return(-1);
}

char *gettagstr(dst, src)
char *dst;
char *src;
{
  int i;
  for(i=0;i<strlen(src);i++)
  {
    if(((src[i] >= 'a') && (src[i] <= 'z')) ||
       ((src[i] >= 'A') && (src[i] <= 'Z')) ||
       ((src[i] >= '1') && (src[i] <= '6')))
      *dst++ = src[i] | 0x20; /* to upper case */
    else
      break;
  }
  *dst = '\0';
  return(&src[i]);
}

int checkqt(p1, p2)
char *p1, *p2;
{
  char *p, qt;
  int cnt = 0;

  /* find first quotation */
  p = p1;
next:
  for(; p < p2; p++)
  {
    if(*p == '"' || *p == '\'')
    {
      qt = *p;
      cnt ++;
      p++;
      break;
    }
  }
  if(cnt == 0) return(0);
  for(;p < p2; p++)
  {
    if(*p == qt)
    {
      cnt++;
      p++;
      goto next;
    }
  }
  if(cnt&1 == 1)
    return(-1);
  return(0);
}

void htmlout(str, taglevel)
char *str;
int taglevel;
{
  char *p, *p1, *p2, tag[256];
  int i, t;
  char *lstr = (char *)malloc(strlen(str)*2);

  tags = 0;

  if(CONV == 1)
  {
    kana2roma(lstr, str, strlen(str)*2);
    p = lstr;
  }
  else
  {
    p = str;
  }
  while(1)
  {
    if((p1 = (char *)strchr(p, '<')) == NULL)
    {
      printf("%s", p);
      break;
    }
    for(p2 = p; p2 < p1; p2++)
      putchar(*p2);
    p1++;
    if(*p1 != '/') /* begin of tag (<tag>) */
    {
      char *p3;
      p2 = gettagstr(tag, p1);
      if((p3 = (char *)strchr(p2, '>')) != NULL)
      {
        if(checkqt(p2, p3) == 0)
          t = isvalidtag(tag);
        else
          t = -1;
      }
      else
        t = -1;
      if(t == -1) /* unsupported tag */
      {
        printf("&lt;");
      }
      else
      {
        if(t > 0)
          tagstack[tags++] = t;
        printf("<");
        for(;p1<=p3;p1++)
          putchar(*p1);
        if(t == 0 && strcmp(tag, "br") == 0)
          printf("&gt;");
      }
      p = p1;
    }
    else /* end of tag (</tag>) */
    {
      p2 = gettagstr(tag, p1 + 1);
      if(strchr(p2, '>') != NULL)
        t = isvalidtag(tag);
      else
        t = -1;
      if(t > 0)
      {
        int i;
        for(i = tags - 1; i >= 0; i--)
        {
          if(tagstack[i] == t)
          {
            tagstack[i] = 0;
            break;
          }
        }
      }
      if(t == -1)
        printf("&lt;");
      else
        printf("<");
      p = p1;
    }
  }
  for(i = tags - 1; i >= 0; i--)
  {
    if(tagstack[i] > 0)
    {
      printf("</%s>", validtags2[tagstack[i] - 1]);
    }
  }
  free(lstr);
}


void nohtmlout(str)
char *str;
{
  int i, tag = 0;
  char *p, *lstr = (char *)malloc(strlen(str)*2);
  if(CONV == 1)
  {
    kana2roma(lstr, str, strlen(str)*2);
    p = lstr;
  }
  else
  {
    p = str;
  }
  for(i = 0;i < strlen(p); i++)
  {
    if(p[i] == '<') {tag = 1; continue;}
    if(p[i] == '>') {tag = 0; continue;}
    if(tag == 0) putchar(p[i]);
  }
  free(lstr);
}

void body_out(body)
BODY *body;
{
  printf("<BODY");
  if(body->bgcolor != NULL && *body->bgcolor != '\0')
    printf(" BGCOLOR=\"%s\"", body->bgcolor);
  else if(body->bgimage != NULL && *body->bgimage != '\0')
    printf(" BACKGROUND=\"%s\"", body->bgimage);
  if(body->txcolor != NULL && *body->txcolor != '\0')
    printf(" TEXT=\"%s\"", body->txcolor);
  if(body->licolor != NULL && *body->licolor != '\0')
    printf(" LINK=\"%s\"", body->licolor);
  if(body->vlcolor != NULL && *body->vlcolor != '\0')
    printf(" VLINK=\"%s\"", body->vlcolor);
  if(body->alcolor != NULL && *body->alcolor != '\0')
    printf(" ALINK=\"%s\"", body->alcolor);
  printf(">\n");
}

void print_string(str)
char *str;
{
  while(*str != '\0')
  {
    if(*str == '"')
      printf("&quot;");
    else if(*str == '&')
      printf("&amp;");
    else
      putchar(*str);
    str++;
  }
}

void hidden_out(name, value)
char *name;
char *value;
{
  if(*value != '\0')
  {
    printf("<INPUT TYPE=\"HIDDEN\" NAME=\"");
    print_string(name);
    printf("\" VALUE=\"");
    print_string(value);
    printf("\">\n");
  }
}

struct sccolor {
  char *color;
  int red;
  int green;
  int blue;
};

struct sccolor validcolors[] = {
  { "aliceblue", 240, 248, 255 },
  { "antiquewhite", 250, 235, 215 },
  { "aquamarine", 127, 255, 212 },
  { "azure", 240, 255, 255 },
  { "beige", 245, 245, 220 },
  { "bisque", 255, 228, 196 },
  { "black", 0, 0, 0 },
  { "blanchedalmond", 255, 235, 205 },
  { "blue", 0, 0, 255 },
  { "blueviolet", 138, 43, 226 },
  { "brown", 165, 42, 42 },
  { "burlywood", 222, 184, 135 },
  { "cadetblue", 95, 158, 160 },
  { "chartreuse", 127, 255, 0 },
  { "chocolate", 210, 105, 30 },
  { "coral", 255, 127, 80 },
  { "cornflowerblue", 100, 149, 237 },
  { "cornsilk", 255, 248, 220 },
  { "cyan", 0, 255, 255 },
  { "darkblue", 0, 0, 139 },
  { "darkcyan", 0, 139, 139 },
  { "darkgoldenrod", 184, 134, 11 },
  { "darkgray", 169, 169, 169 },
  { "darkgreen", 0, 100, 0 },
  { "darkgrey", 169, 169, 169 },
  { "darkkhaki", 189, 183, 107 },
  { "darkmagenta", 139, 0, 139 },
  { "darkolivegreen", 85, 107, 47 },
  { "darkorange", 255, 140, 0 },
  { "darkorchid", 153, 50, 204 },
  { "darkred", 139, 0, 0 },
  { "darksalmon", 233, 150, 122 },
  { "darkseagreen", 143, 188, 143 },
  { "darkslateblue", 72, 61, 139 },
  { "darkslategray", 47, 79, 79 },
  { "darkslategrey", 47, 79, 79 },
  { "darkturquoise", 0, 206, 209 },
  { "darkviolet", 148, 0, 211 },
  { "deeppink", 255, 20, 147 },
  { "deepskyblue", 0, 191, 255 },
  { "dimgray", 105, 105, 105 },
  { "dimgrey", 105, 105, 105 },
  { "dodgerblue", 30, 144, 255 },
  { "firebrick", 178, 34, 34 },
  { "floralwhite", 255, 250, 240 },
  { "forestgreen", 34, 139, 34 },
  { "gainsboro", 220, 220, 220 },
  { "ghostwhite", 248, 248, 255 },
  { "gold", 255, 215, 0 },
  { "goldenrod", 218, 165, 32 },
  { "gray", 190, 190, 190 },
  { "green", 0, 255, 0 },
  { "greenyellow", 173, 255, 47 },
  { "grey", 190, 190, 190 },
  { "honeydew", 240, 255, 240 },
  { "hotpink", 255, 105, 180 },
  { "indianred", 205, 92, 92 },
  { "ivory", 255, 255, 240 },
  { "khaki", 240, 230, 140 },
  { "lavender", 230, 230, 250 },
  { "lavenderblush", 255, 240, 245 },
  { "lawngreen", 124, 252, 0 },
  { "lemonchiffon", 255, 250, 205 },
  { "lightblue", 173, 216, 230 },
  { "lightcoral", 240, 128, 128 },
  { "lightcyan", 224, 255, 255 },
  { "lightgoldenrod", 238, 221, 130 },
  { "lightgoldenrodyellow", 250, 250, 210 },
  { "lightgray", 211, 211, 211 },
  { "lightgreen", 144, 238, 144 },
  { "lightgrey", 211, 211, 211 },
  { "lightpink", 255, 182, 193 },
  { "lightsalmon", 255, 160, 122 },
  { "lightseagreen", 32, 178, 170 },
  { "lightskyblue", 135, 206, 250 },
  { "lightslateblue", 132, 112, 255 },
  { "lightslategray", 119, 136, 153 },
  { "lightslategrey", 119, 136, 153 },
  { "lightsteelblue", 176, 196, 222 },
  { "lightyellow", 255, 255, 224 },
  { "limegreen", 50, 205, 50 },
  { "linen", 250, 240, 230 },
  { "magenta", 255, 0, 255 },
  { "maroon", 176, 48, 96 },
  { "mediumaquamarine", 102, 205, 170 },
  { "mediumblue", 0, 0, 205 },
  { "mediumorchid", 186, 85, 211 },
  { "mediumpurple", 147, 112, 219 },
  { "mediumseagreen", 60, 179, 113 },
  { "mediumslateblue", 123, 104, 238 },
  { "mediumspringgreen", 0, 250, 154 },
  { "mediumturquoise", 72, 209, 204 },
  { "mediumvioletred", 199, 21, 133 },
  { "midnightblue", 25, 25, 112 },
  { "mintcream", 245, 255, 250 },
  { "mistyrose", 255, 228, 225 },
  { "moccasin", 255, 228, 181 },
  { "navajowhite", 255, 222, 173 },
  { "navy", 0, 0, 128 },
  { "navyblue", 0, 0, 128 },
  { "oldlace", 253, 245, 230 },
  { "olivedrab", 107, 142, 35 },
  { "orange", 255, 165, 0 },
  { "orangered", 255, 69, 0 },
  { "orchid", 218, 112, 214 },
  { "palegoldenrod", 238, 232, 170 },
  { "palegreen", 152, 251, 152 },
  { "paleturquoise", 175, 238, 238 },
  { "palevioletred", 219, 112, 147 },
  { "papayawhip", 255, 239, 213 },
  { "peachpuff", 255, 218, 185 },
  { "peru", 205, 133, 63 },
  { "pink", 255, 192, 203 },
  { "plum", 221, 160, 221 },
  { "powderblue", 176, 224, 230 },
  { "purple", 160, 32, 240 },
  { "red", 255, 0, 0 },
  { "rosybrown", 188, 143, 143 },
  { "royalblue", 65, 105, 225 },
  { "saddlebrown", 139, 69, 19 },
  { "salmon", 250, 128, 114 },
  { "sandybrown", 244, 164, 96 },
  { "seagreen", 46, 139, 87 },
  { "seashell", 255, 245, 238 },
  { "sienna", 160, 82, 45 },
  { "skyblue", 135, 206, 235 },
  { "slateblue", 106, 90, 205 },
  { "slategray", 112, 128, 144 },
  { "slategrey", 112, 128, 144 },
  { "snow", 255, 250, 250 },
  { "springgreen", 0, 255, 127 },
  { "steelblue", 70, 130, 180 },
  { "tan", 210, 180, 140 },
  { "thistle", 216, 191, 216 },
  { "tomato", 255, 99, 71 },
  { "turquoise", 64, 224, 208 },
  { "violet", 238, 130, 238 },
  { "violetred", 208, 32, 144 },
  { "wheat", 245, 222, 179 },
  { "white", 255, 255, 255 },
  { "whitesmoke", 245, 245, 245 },
  { "yellow", 255, 255, 0 },
  { "yellowgreen", 154, 205, 50 },
  { NULL, 0, 0, 0}
};

int getrgb(red, green, blue, str)
int *red, *green, *blue;
char *str;
{
  char buf[256];
  int i;

  if(*str == '#' || *str >= '0' && *str <= '9')
  {
    if(*str == '#') str++;
    if(strlen(str) != 6) return(0);
    strncpy(buf, str, 2); buf[2] = ' ';
    strncpy(&buf[3], str+2, 2); buf[5] = ' ';
    strncpy(&buf[6], str+4, 2); buf[8] = '\0';
    if(sscanf(buf, "%x %x %x", red, green, blue) < 3) return(0);
    return(1);
  }
  else
  {
    strncpy(buf, str, sizeof(buf));
    scstrlwr(buf);
    for(i = 0;;i++)
    {
      int r;
      if(validcolors[i].color == NULL) return(0);
      r = strcmp(buf, validcolors[i].color);
      if(r < 0) return(0);
      else if(r == 0)
      {
        *red = validcolors[i].red;
        *green = validcolors[i].green;
        *blue = validcolors[i].blue;
        return(1);
      }
    }
  }
  return(0);
}

int scabs(a)
int a;
{
  if(a < 0) return(-a);
  return(a);
}

int isvalidcolor(color, incolor)
char *color;
char *incolor;
{
  int r1, g1, b1, r2, g2, b2;
  if(getrgb(&r1, &g1, &b1, color) == 0) return(0);
  if(getrgb(&r2, &g2, &b2, incolor) == 0) return(0);
  if((scabs(r1 - r2) + scabs(g1 - g2) + scabs(b1 - b2)) < 32*3)
    return(0);
  return(1);
}

